import argshell


def get_edit_config_parser() -> argshell.ArgShellParser:
    parser = argshell.ArgShellParser(
        "config",
        description="Edit or create the `hassle_config.toml` file.",
    )
    parser.add_argument(
        "-n",
        "--name",
        type=str,
        default=None,
        help=" Your name. This will be used to populate the 'authors' field of a packages 'pyproject.toml'. ",
    )
    parser.add_argument(
        "-e",
        "--email",
        type=str,
        default=None,
        help=" Your email. This will be used to populate the 'authors' field of a packages 'pyproject.toml'. ",
    )
    parser.add_argument(
        "-g",
        "--github_username",
        type=str,
        default=None,
        help=""" Your github username name. 
        When creating a new package, say with the name 'mypackage', the pyproject.toml 'Homepage' field  will be set to 
        'https://github.com/{github_username}/mypackage' 
        and the 'Source code' field will be set to 
        'https://github.com/{github_username}/mypackage/tree/main/src/mypackage'.""",
    )
    parser.add_argument(
        "-d",
        "--docs_url",
        type=str,
        default=None,
        help=""" The template url to be used in your pyproject.toml file indicating where your project docs will be hosted.
        Pass the url with '$name' as a placeholder for where the package name should go, 
        e.g. 'https://somedocswebsite/user/projects/$name'.
        If 'hassle_config.toml' didn't exist prior to running this tool and nothing is given for this arg, it will default to using the package's github url. 
        e.g. for package 'mypackage' the url will be 'https://github.com/{github_username}/mypackage/tree/main/docs' """,
    )
    parser.add_argument(
        "-t",
        "--tag_prefix",
        type=str,
        default=None,
        help=""" When using Hassle to do `git tag`, this will be prefixed to the front of the version number in the `pyproject.toml` file.""",
    )
    return parser


def get_new_project_parser() -> argshell.ArgShellParser:
    parser = argshell.ArgShellParser(
        "new", description="Create a new project in the current directory."
    )

    parser.add_argument(
        "name",
        type=str,
        help=""" Name of the package to create in the current working directory. """,
    )

    parser.add_argument(
        "-s",
        "--source_files",
        nargs="*",
        type=str,
        default=[],
        help=""" List of additional source files to create in addition to the default
        __init__.py and {name}.py files.""",
    )

    parser.add_argument(
        "-d",
        "--description",
        type=str,
        default="",
        help=""" The package description to be added to the pyproject.toml file. """,
    )

    parser.add_argument(
        "-dp",
        "--dependencies",
        nargs="*",
        type=str,
        default=[],
        help=""" List of dependencies to add to pyproject.toml.
        Note: hassle.py will automatically scan your project for 3rd party imports and update pyproject.toml. 
        This switch is largely useful for adding dependencies your project might need, 
        but doesn't directly import in any source files, like an os.system() call that invokes a 3rd party cli.""",
    )

    parser.add_argument(
        "-k",
        "--keywords",
        nargs="*",
        type=str,
        default=[],
        help=""" List of keywords to be added to the keywords field in pyproject.toml. """,
    )

    parser.add_argument(
        "-as",
        "--add_script",
        action="store_true",
        help=""" Add section to pyproject.toml declaring the package should be installed with command line scripts added. 
        The default is '{package_name} = "{package_name}.{package_name}:main".""",
    )

    parser.add_argument(
        "-nl",
        "--no_license",
        action="store_true",
        help=""" By default, projects are created with an MIT license.
        Set this flag to avoid adding a license if you want to configure licensing at another time.""",
    )

    parser.add_argument(
        "-np",
        "--not_package",
        action="store_true",
        help=""" Put source files in top level directory and delete tests folder. """,
    )
    return parser


def get_build_parser() -> argshell.ArgShellParser:
    """Returns a build parser."""
    parser = argshell.ArgShellParser(
        "build",
        description=""" Run the build process:
         * Run tests (abandoning build on failure)
         * Sort imports and format with Black
         * Update dependencies
         * Generate docs
         * Invoke `build` module""",
    )
    parser.add_argument(
        "-s", "--skip_tests", action="store_true", help=""" Skip running tests. """
    )
    parser.add_argument(
        "-o",
        "--overwrite_dependencies",
        action="store_true",
        help=""" Overwrite dependencies instead of appending new ones to the current contents of `pyproject.toml`. """,
    )
    parser.add_argument(
        "-v",
        "--include_versions",
        action="store_true",
        help=""" Include versions when adding dependencies. """,
    )
    return parser


def get_update_parser() -> argshell.ArgShellParser:
    """Returns a update parser."""
    parser = get_build_parser()
    parser.prog = "update"
    parser.description = """
    Update this package:
    * Run the build command (run `hassle build -h` for info)
    * Increment project version
    * Update/create changelog
    * git commit -m "chore: build {version}"
    * Git tag with new version number
    * Pull/push to remote
    """
    parser.add_argument(
        "update_type",
        type=str,
        choices=("major", "minor", "patch"),
        help=""" The type of update. """,
    )
    parser.add_argument(
        "-p",
        "--publish",
        action="store_true",
        help=""" Publish theupdated package to PYPI. """,
    )
    parser.add_argument(
        "-i",
        "--install",
        action="store_true",
        help=""" Install updated package. """,
    )
    return parser


def get_add_script_parser() -> argshell.ArgShellParser:
    """Returns a add_script parser."""
    parser = argshell.ArgShellParser(
        "add_script", description=""" Add a script to the `pyproject.toml` file. """
    )
    parser.add_argument("name", type=str, help=""" The name of the script """)
    parser.add_argument(
        "file", type=str, help=""" The file stem the function/script is located in. """
    )
    parser.add_argument(
        "function",
        nargs="?",
        default="main",
        help=""" The name of the function to invoke with the script. Defaults to `main`. """,
    )
    return parser


def add_default_source_files(args: argshell.Namespace) -> argshell.Namespace:
    """Add `__init__.py` and `{args.name}.py` to `args.source_files`."""
    args.source_files += ["__init__.py", f"{args.name}.py"]
    return args



import re
import subprocess
from typing import Any

import requests
from bs4 import BeautifulSoup
from gitbetter import Git
from pathier import Pathier

root = Pathier(__file__).parent


def swap_keys(data: dict[str, Any], keys: tuple[str, str]) -> dict[str, Any]:
    """Convert between keys in `data`.
    The order of `keys` doesn't matter.
    >>> data = {"one two": 1}
    >>> data = swap_keys(data, ("one two", "one-two"))
    >>> print(data)
    >>> {"one-two": 1}
    >>> data = swap_keys(data, ("one two", "one-two"))
    >>> print(data)
    >>> {"one two": 1}
    """
    key1, key2 = keys
    data_keys = data.keys()
    if key1 in data_keys:
        data[key2] = data.pop(key1)
    elif key2 in data_keys:
        data[key1] = data.pop(key2)
    return data


def run_tests(pytest_args: list[str] = []) -> bool:
    """Invoke `coverage` and `pytest -s`.

    Returns `True` if all tests passed or if no tests were found."""
    results = subprocess.run(["coverage", "run", "-m", "pytest", "-s"] + pytest_args)
    subprocess.run(["coverage", "report", f"--include={Pathier.cwd()}/*"])
    subprocess.run(["coverage", "html", f"--include={Pathier.cwd()}/*"])
    return results.returncode in [0, 5]


def check_pypi(package_name: str) -> bool:
    """Check if a package with package_name already exists on `pypi.org`.
    Returns `True` if package name exists.
    Only checks the first page of results."""
    url = f"https://pypi.org/search/?q={package_name.lower()}"
    response = requests.get(url)
    if response.status_code != 200:
        raise RuntimeError(
            f"Error: pypi.org returned status code: {response.status_code}"
        )
    soup = BeautifulSoup(response.text, "html.parser")
    pypi_packages = [
        span.text.lower()
        for span in soup.find_all("span", class_="package-snippet__name")
    ]
    return package_name in pypi_packages


def get_answer(question: str) -> bool | None:
    """Repeatedly ask the user a yes/no question until a 'y' or a 'n' is received."""
    ans = ""
    question = question.strip()
    if "?" not in question:
        question += "?"
    question += " (y/n): "
    while ans not in ["y", "yes", "no", "n"]:
        ans = input(question).strip().lower()
        if ans in ["y", "yes"]:
            return True
        elif ans in ["n", "no"]:
            return False
        else:
            print("Invalid answer.")


def bump_version(current_version: str, bump_type: str) -> str:
    """Bump `current_version` according to `bump_type` and return the new version.

    #### :params:

    `current_version`: A version string conforming to Semantic Versioning standards.
    i.e. `{major}.{minor}.{patch}`

    `bump_type` can be one of `major`, `minor`, or `patch`.

    Raises an exception if `current_version` is formatted incorrectly or if `bump_type` isn't one of the aforementioned types.
    """
    if not re.findall(r"[0-9]+.[0-9]+.[0-9]+", current_version):
        raise ValueError(
            f"{current_version} does not appear to match the required format of `x.x.x`."
        )
    bump_type = bump_type.lower().strip()
    if bump_type not in ["major", "minor", "patch"]:
        raise ValueError(
            f"`bump_type` {bump_type} is not one of `major`, `minor`, or `patch`."
        )
    major, minor, patch = [int(part) for part in current_version.split(".")]
    if bump_type == "major":
        major += 1
        minor = 0
        patch = 0
    elif bump_type == "minor":
        minor += 1
        patch = 0
    elif bump_type == "patch":
        patch += 1
    return f"{major}.{minor}.{patch}"


def on_primary_branch() -> bool:
    """Returns `False` if repo is not currently on `main` or `master` branch."""
    git = Git(True)
    if git.current_branch not in ["main", "master"]:
        return False
    return True



import subprocess
from dataclasses import asdict, dataclass, field
from datetime import datetime
from functools import cached_property
from typing import Any

import black
import dacite
import isort
import requests
from bs4 import BeautifulSoup, Tag
from packagelister import packagelister
from pathier import Pathier, Pathish
from typing_extensions import Self

from hassle import utilities

root = Pathier(__file__).parent


@dataclass
class Sdist:
    exclude: list[str]


@dataclass
class Targets:
    sdist: Sdist


@dataclass
class Build:
    targets: Targets


@dataclass
class BuildSystem:
    requires: list[str]
    build_backend: str


@dataclass
class Urls:
    Homepage: str = ""
    Documentation: str = ""
    Source_code: str = ""


@dataclass
class Author:
    name: str = ""
    email: str = ""


@dataclass
class Git:
    tag_prefix: str = ""


@dataclass
class IniOptions:
    addopts: list[str]
    pythonpath: str


@dataclass
class Pytest:
    ini_options: IniOptions


@dataclass
class Hatch:
    build: Build


@dataclass
class Tool:
    pytest: Pytest
    hatch: Hatch


@dataclass
class Project:
    name: str
    authors: list[Author] = field(default_factory=list)
    description: str = ""
    requires_python: str = ""
    version: str = ""
    dependencies: list[str] = field(default_factory=list)
    readme: str = ""
    keywords: list[str] = field(default_factory=list)
    classifiers: list[str] = field(default_factory=list)
    urls: Urls = field(default_factory=Urls)
    scripts: dict[str, str] = field(default_factory=dict)


@dataclass
class Pyproject:
    build_system: BuildSystem
    project: Project
    tool: Tool

    @staticmethod
    def _swap_keys(data: dict[str, Any]) -> dict[str, Any]:
        """Swap between original toml key and valid Python variable."""
        if "build-system" in data:
            data = utilities.swap_keys(data, ("build-system", "build_system"))
            if "build-backend" in data["build_system"]:
                data["build_system"] = utilities.swap_keys(
                    data["build_system"], ("build-backend", "build_backend")
                )
        elif "build_system" in data:
            data = utilities.swap_keys(data, ("build-system", "build_system"))
            if "build_backend" in data["build-system"]:
                data["build-system"] = utilities.swap_keys(
                    data["build-system"], ("build-backend", "build_backend")
                )

        if "project" in data and (
            "requires-python" in data["project"] or "requires_python"
        ):
            data["project"] = utilities.swap_keys(
                data["project"], ("requires-python", "requires_python")
            )
        if all(
            [
                "project" in data,
                "urls" in data["project"],
                (
                    "Source code" in data["project"]["urls"]
                    or "Source_code" in data["project"]["urls"]
                ),
            ]
        ):
            data["project"]["urls"] = utilities.swap_keys(
                data["project"]["urls"], ("Source code", "Source_code")
            )

        return data

    @classmethod
    def load(cls, path: Pathish = Pathier("pyproject.toml")) -> Self:
        """Return a `datamodel` object populated from `path`."""
        data = Pathier(path).loads()
        data = cls._swap_keys(data)
        return dacite.from_dict(cls, data)

    def dump(self, path: Pathish = Pathier("pyproject.toml")):
        """Write the contents of this `datamodel` object to `path`."""
        data = asdict(self)
        data = self._swap_keys(data)
        Pathier(path).dumps(data)

    @classmethod
    def from_template(cls) -> Self:
        """Return a `Pyproject` object using `templates/pyproject_template.toml`."""
        return cls.load(root / "templates" / "pyproject.toml")


@dataclass
class HassleConfig:
    authors: list[Author] = field(default_factory=list)
    project_urls: Urls = field(default_factory=Urls)
    git: Git = field(default_factory=Git)

    @classmethod
    def load(
        cls, path: Pathish = Pathier(__file__).parent / "hassle_config.toml"
    ) -> Self:
        """Return a `datamodel` object populated from `path`."""
        path = Pathier(path)
        if not path.exists():
            raise FileNotFoundError(
                f"Could not find hassle config at {path}.\nRun hassle_config in a terminal to set it."
            )
        data = path.loads()
        data["project_urls"] = utilities.swap_keys(
            data["project_urls"], ("Source_code", "Source code")
        )
        return dacite.from_dict(cls, data)

    def dump(self, path: Pathish = Pathier(__file__).parent / "hassle_config.toml"):
        """Write the contents of this `datamodel` object to `path`."""
        data = asdict(self)
        data["project_urls"] = utilities.swap_keys(
            data["project_urls"], ("Source_code", "Source code")
        )
        Pathier(path).dumps(data)

    @staticmethod
    def warn():
        print("hassle_config.toml has not been set.")
        print("Run hassle_config to set it.")
        print("Run 'hassle config -h' for help.")

    @staticmethod
    def exists(path: Pathish = Pathier(__file__).parent / "hassle_config.toml") -> bool:
        return Pathier(path).exists()

    @classmethod
    def configure(
        cls,
        name: str | None = None,
        email: str | None = None,
        github_username: str | None = None,
        docs_url: str | None = None,
        tag_prefix: str | None = None,
        config_path: Pathish = Pathier(__file__).parent / "hassle_config.toml",
    ):
        """Create or edit `hassle_config.toml` from given params."""
        print(f"Manual edits can be made at {config_path}")
        if not cls.exists(config_path):
            config = cls()
        else:
            config = cls.load(config_path)
        # Add an author to config if a name or email is given.
        if name or email:
            config.authors.append(Author(name or "", email or ""))
        if github_username:
            homepage = f"https://github.com/{github_username}/$name"
            config.project_urls.Homepage = homepage
            config.project_urls.Source_code = f"{homepage}/tree/main/src/$name"
        if not config.project_urls.Documentation:
            if github_username and not docs_url:
                config.project_urls.Documentation = (
                    f"https://github.com/{github_username}/$name/tree/main/docs"
                )
            elif docs_url:
                config.project_urls.Documentation = docs_url
        if tag_prefix:
            config.git.tag_prefix = tag_prefix
        config.dump(config_path)


@dataclass
class HassleProject:
    pyproject: Pyproject
    projectdir: Pathier
    source_files: list[str]
    templatedir: Pathier = root / "templates"

    @property
    def source_code(self) -> str:
        """Join and return all code from any `.py` files in `self.srcdir`.

        Useful if a tool needs to scan all the source code for something."""
        return "\n".join(file.read_text() for file in self.srcdir.rglob("*.py"))

    @cached_property
    def srcdir(self) -> Pathier:
        return self.projectdir / "src" / self.pyproject.project.name

    @cached_property
    def changelog_path(self) -> Pathier:
        return self.projectdir / "CHANGELOG.md"

    @cached_property
    def pyproject_path(self) -> Pathier:
        return self.projectdir / "pyproject.toml"

    @cached_property
    def docsdir(self) -> Pathier:
        return self.projectdir / "docs"

    @cached_property
    def testsdir(self) -> Pathier:
        return self.projectdir / "tests"

    @cached_property
    def vsdir(self) -> Pathier:
        return self.projectdir / ".vscode"

    @cached_property
    def distdir(self) -> Pathier:
        return self.projectdir / "dist"

    @property
    def name(self) -> str:
        """This package's name."""
        return self.pyproject.project.name

    @property
    def version(self) -> str:
        """This package's version."""
        return self.pyproject.project.version

    @version.setter
    def version(self, new_version: str):
        self.pyproject.project.version = new_version

    @classmethod
    def load(cls, projectdir: Pathish) -> Self:
        """Load a project given `projectdir`."""
        projectdir = Pathier(projectdir)
        pyproject = Pyproject.load(projectdir / "pyproject.toml")
        name = pyproject.project.name
        # Convert source files to path stems relative to projectdir/src/name
        # e.g `C:/python/projects/hassle/src/hassle/templates/pyproject.toml`
        # becomes `templates/pyproject.toml`
        source_files = [
            str(file.separate(name))
            for file in (projectdir / "src" / name).rglob("*")
            if file.is_file()
        ]
        return cls(pyproject, projectdir, source_files)

    @classmethod
    def new(
        cls,
        targetdir: Pathier,
        name: str,
        description: str = "",
        dependencies: list[str] = [],
        keywords: list[str] = [],
        source_files: list[str] = [],
        add_script: bool = False,
        no_license: bool = False,
    ) -> Self:
        """Create and return a new hassle project."""
        pyproject = Pyproject.from_template()
        config = HassleConfig.load()
        pyproject.project.name = name
        pyproject.project.authors = config.authors
        pyproject.project.description = description
        pyproject.project.dependencies = dependencies
        pyproject.project.keywords = keywords
        pyproject.project.urls.Homepage = config.project_urls.Homepage.replace(
            "$name", name
        )
        pyproject.project.urls.Documentation = (
            config.project_urls.Documentation.replace("$name", name)
        )
        pyproject.project.urls.Source_code = config.project_urls.Source_code.replace(
            "$name", name
        )
        hassle = cls(pyproject, targetdir, source_files)
        if add_script:
            hassle.add_script(name, name)
        hassle.generate_files()
        if no_license:
            hassle.pyproject.project.classifiers.pop(1)
            (hassle.projectdir / "LICENSE.txt").delete()
        hassle.save()
        return hassle

    def get_template(self, file_name: str) -> str:
        """Open are return the content of `{self.templatedir}/{file_name}`."""
        return (self.templatedir / file_name).read_text()

    def save(self):
        """Dump `self.pyproject` to `{self.projectdir}/pyproject.toml`."""
        self.pyproject.dump(self.pyproject_path)

    def format_source_files(self):
        """Use isort and black to format files"""
        for file in self.projectdir.rglob("*.py"):
            isort.file(file)
        try:
            black.main([str(self.projectdir)])
        except SystemExit as e:
            ...
        except Exception as e:
            raise e

    def latest_version_is_published(self) -> bool:
        """Check if the current version of this project has been published to pypi.org."""
        pypi_url = f"https://pypi.org/project/{self.name}"
        response = requests.get(pypi_url)
        if response.status_code != 200:
            raise RuntimeError(
                f"{pypi_url} returned status code {response.status_code} :/"
            )
        soup = BeautifulSoup(response.text, "html.parser")
        header = soup.find("h1", class_="package-header__name")
        assert isinstance(header, Tag)
        text = header.text.strip()
        pypi_version = text[text.rfind(" ") + 1 :]
        return self.version == pypi_version

    # ====================================================================================
    # Updaters ===========================================================================
    # ====================================================================================
    def add_script(self, name: str, file_stem: str, function: str = "main"):
        """Add a script to `pyproject.project.scripts` in the format `{name} = "{package_name}.{file_stem}:{function}"`"""
        self.pyproject.project.scripts[name] = f"{self.name}.{file_stem}:{function}"

    def update_init_version(self):
        """Update the `__version__` in this projects `__init__.py` file
        to the current value of `self.pyproject.project.version`
        if it exists and has a `__version__` string.

        If it doesn't have a `__version__` string, append one to it."""
        init_file = self.srcdir / "__init__.py"
        version = f'__version__ = "{self.version}"'
        if init_file.exists():
            content = init_file.read_text()
            if "__version__" in content:
                lines = content.splitlines()
                for i, line in enumerate(lines):
                    if line.startswith("__version__"):
                        lines[i] = version
                content = "\n".join(lines)
            else:
                content += f"\n{version}"
            init_file.write_text(content)

    def bump_version(self, bump_type: str):
        """Bump the version of this project.

        `bump_type` should be `major`, `minor`, or `patch`."""
        # bump pyproject version
        self.version = utilities.bump_version(self.version, bump_type)
        # bump `__version__` in __init__.py if the file exists and has a `__version__`.
        self.update_init_version()

    def update_dependencies(
        self, overwrite_existing_packages: bool, include_versions: bool
    ):
        """Scan project for dependencies and update the corresponding field in the pyproject model.

        If `overwrite_existing_packages` is `False`, this function will only add a package if it isn't already listed,
        but won't remove anything currently in the list.
        Use this option to preserve manually added dependencies."""
        project = packagelister.scan_dir(self.srcdir)
        version_conditional = ">=" if include_versions else None
        if overwrite_existing_packages:
            self.pyproject.project.dependencies = project.get_formatted_requirements(
                version_conditional
            )
        else:
            # Only add a package if it isn't already in the dependency list
            self.pyproject.project.dependencies.extend(
                [
                    (
                        package.get_formatted_requirement(version_conditional)
                        if version_conditional
                        else package.distribution_name
                    )
                    for package in project.requirements
                    if all(
                        package.distribution_name not in existing_dependency
                        for existing_dependency in self.pyproject.project.dependencies
                    )
                ]
            )

    def _generate_changelog(self) -> list[str]:
        if HassleConfig.exists():
            tag_prefix = HassleConfig.load().git.tag_prefix
        else:
            HassleConfig.warn()
            print("Assuming no tag prefix.")
            tag_prefix = ""
        raw_changelog = [
            line
            for line in subprocess.run(
                [
                    "auto-changelog",
                    "-p",
                    self.projectdir,
                    "--tag-prefix",
                    tag_prefix,
                    "--stdout",
                ],
                stdout=subprocess.PIPE,
                text=True,
            ).stdout.splitlines(True)
            if not line.startswith(
                (
                    "Full set of changes:",
                    f"* build {tag_prefix}",
                    "* update changelog",
                )
            )
        ]
        return raw_changelog

    def update_changelog(self):
        """Update `CHANGELOG.md` by invoking the `auto-changelog` module.

        If `hassle_config.toml` doesn't exist, an empty tag prefix will be assumed."""
        raw_changelog = self._generate_changelog()
        # If there's no existing changelog, dump the generated one and get out of here.
        if not self.changelog_path.exists():
            self.changelog_path.join(raw_changelog)
            return

        # Don't want to overwrite previously existing manual changes/edits
        existing_changelog = self.changelog_path.read_text().splitlines(True)[
            2:
        ]  # First two elements are "# Changelog\n" and "\n"
        new_changes = raw_changelog
        for line in existing_changelog:
            # Release headers are prefixed with "## "
            if line.startswith("## "):
                new_changes = raw_changelog[: raw_changelog.index(line)]
                break
        changes = "".join(new_changes)
        # "#### OTHERS" gets added to the changelog even when there's nothing for that category,
        # so we'll get rid of it if that's the case
        others = "#### Others"
        if changes.strip("\n").endswith(others):
            changes = changes.strip("\n").replace(others, "\n\n")
        # If changes == "# Changelog\n\n" then there weren't actually any new changes
        if not changes == "# Changelog\n\n":
            self.changelog_path.write_text(changes + "".join(existing_changelog))

    # ====================================================================================
    # File/Project creation ==============================================================
    # ====================================================================================

    def create_source_files(self):
        """Generate source files in `self.srcdir`."""
        for file in self.source_files:
            (self.srcdir / file).touch()
        init = self.srcdir / "__init__.py"
        if init.exists():
            init.append(f'__version__ = "{self.version}"')

    def create_readme(self):
        readme = self.get_template("README.md")
        readme = readme.replace("$name", self.name)
        readme = readme.replace("$description", self.pyproject.project.description)
        (self.projectdir / "README.md").write_text(readme)

    def create_license(self):
        license_ = self.get_template("license.txt")
        license_ = license_.replace("$year", str(datetime.now().year))
        (self.projectdir / "LICENSE.txt").write_text(license_)

    def create_gitignore(self):
        (self.templatedir / ".gitignore.txt").copy(self.projectdir / ".gitignore")

    def create_vscode_settings(self):
        self.vsdir.mkdir()
        (self.templatedir / "vscode_settings.json").copy(self.vsdir / "settings.json")

    def create_tests(self):
        (self.testsdir / f"test_{self.name}.py").touch()

    def create_py_typed(self):
        (self.projectdir / "py.typed").touch()

    def generate_files(self):
        """Create all the necessary files.

        Note: This will overwrite any existing files."""
        self.projectdir.mkdir()
        for func in dir(self):
            if func.startswith("create_"):
                getattr(self, func)()
        self.pyproject.dump(self.pyproject_path)

    def generate_docs(self):
        """Generate docs by invoking `pdoc`"""
        self.docsdir.delete()
        subprocess.run(["pdoc", "-o", self.docsdir, self.srcdir])



import shlex
import subprocess
import sys
from typing import Any

import argshell
import pip
from gitbetter import Git
from pathier import Pathier

from hassle import parsers, utilities
from hassle.models import HassleConfig, HassleProject, Pyproject

root = Pathier(__file__).parent


class HassleShell(argshell.ArgShell):
    def __init__(self, command: str, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        if command == "new":
            # load a blank HassleProject
            self.project = HassleProject(Pyproject.from_template(), Pathier.cwd(), [])
        elif command != "check_pypi":
            try:
                self.project = HassleProject.load(Pathier.cwd())
            except Exception as e:
                print(f"{Pathier.cwd().stem} does not appear to be a Hassle project.")
                print(e)

    def _build(self, args: argshell.Namespace):
        self.project.format_source_files()
        self.project.update_dependencies(
            args.overwrite_dependencies, args.include_versions
        )
        self.project.generate_docs()
        self.project.distdir.delete()
        self.project.save()
        subprocess.run([sys.executable, "-m", "build", Pathier.cwd()])

    @argshell.with_parser(parsers.get_add_script_parser)
    def do_add_script(self, args: argshell.Namespace):
        """Add a script to the `pyproject.toml` file."""
        self.project.add_script(args.name, args.file.strip(".py"), args.function)
        self.project.save()

    @argshell.with_parser(parsers.get_build_parser)
    def do_build(self, args: argshell.Namespace):
        """Build this project."""
        if not args.skip_tests and not utilities.run_tests():
            raise RuntimeError(
                f"ERROR: {Pathier.cwd().stem} failed testing.\nAbandoning build."
            )
        self._build(args)

    def do_check_pypi(self, name: str):
        """Check if the given package name is taken on pypi.org or not."""
        name = name.strip('"')
        if utilities.check_pypi(name):
            print(f"{name} is already taken.")
        else:
            print(f"{name} is available.")

    def do_config(self, _: str = ""):
        """Print hassle config to terminal."""
        config = root / "hassle_config.toml"
        if config.exists():
            print(config.read_text())
        else:
            print("hassle_config.toml doesn't exist.")

    @argshell.with_parser(parsers.get_edit_config_parser)
    def do_configure(self, args: argshell.Namespace):
        """Edit or create `hassle_config.toml`."""
        HassleConfig.configure(
            args.name, args.email, args.github_username, args.docs_url, args.tag_prefix
        )

    def do_format(self, _: str = ""):
        """Format all `.py` files with `isort` and `black`."""
        self.project.format_source_files()

    def do_is_published(self, _: str = ""):
        """Check if the most recent version of this package is published to PYPI."""
        text = f"The most recent version of '{self.project.name}'"
        if self.project.latest_version_is_published():
            print(f"{text} has been published.")
        else:
            print(f"{text} has not been published.")

    @argshell.with_parser(
        parsers.get_new_project_parser,
        [parsers.add_default_source_files],
    )
    def do_new(self, args: argshell.Namespace):
        """Create a new project."""
        # Check if this name is taken.
        if not args.not_package and utilities.check_pypi(args.name):
            print(f"{args.name} already exists on pypi.org")
            if not utilities.get_answer("Continue anyway?"):
                sys.exit()
        # Check if targetdir already exists
        targetdir = Pathier.cwd() / args.name
        if targetdir.exists():
            print(f"'{args.name}' already exists.")
            if not utilities.get_answer("Overwrite?"):
                sys.exit()
        # Load config
        if not HassleConfig.exists():
            HassleConfig.warn()
            if not utilities.get_answer(
                "Continue creating new package with blank config?"
            ):
                raise Exception("Aborting new package creation")
            else:
                print("Creating blank hassle_config.toml...")
                HassleConfig.configure()
        self.project = HassleProject.new(
            targetdir,
            args.name,
            args.description,
            args.dependencies,
            args.keywords,
            args.source_files,
            args.add_script,
            args.no_license,
        )
        # If not a package (just a project) move source code to top level.
        if args.not_package:
            for file in self.project.srcdir.iterdir():
                file.copy(self.project.projectdir / file.name)
            self.project.srcdir.parent.delete()
        # Initialize Git
        self.project.projectdir.mkcwd()
        git = Git()
        git.new_repo()

    def do_publish(self, _: str = ""):
        """Publish this package.

        You must have 'twine' installed and set up to use this command."""
        if not utilities.on_primary_branch():
            print(
                "WARNING: You are trying to publish a project that does not appear to be on its main branch."
            )
            print(f"You are on branch '{Git().current_branch}'")
            if not utilities.get_answer("Continue anyway?"):
                return
        subprocess.run(["twine", "upload", self.project.distdir / "*"])

    def do_test(self, args: str):
        """Invoke `pytest -s` with `Coverage`.

        Optionally, provide any other cli args pytest can accept."""
        utilities.run_tests(shlex.split(args))

    @argshell.with_parser(parsers.get_update_parser)
    def do_update(self, args: argshell.Namespace):
        """Update this package."""
        if not args.skip_tests and not utilities.run_tests():
            raise RuntimeError(
                f"ERROR: {Pathier.cwd().stem} failed testing.\nAbandoning build."
            )
        self.project.bump_version(args.update_type)
        self.project.save()
        self._build(args)
        git = Git()
        if HassleConfig.exists():
            tag_prefix = HassleConfig.load().git.tag_prefix
        else:
            HassleConfig.warn()
            print("Assuming no tag prefix.")
            tag_prefix = ""
        tag = f"{tag_prefix}{self.project.version}"
        git.add_files([self.project.distdir, self.project.docsdir])
        git.add(". -u")
        git.commit(f'-m "chore: build {tag}"')
        # 'auto-changelog' generates based off of commits between tags
        # So to include the changelog in the tagged commit,
        # we have to tag the code, update/commit the changelog, delete the tag, and then retag
        # (One of these days I'll just write my own changelog generator)
        git.tag(tag)
        self.project.update_changelog()
        with git.capturing_output():
            git.tag(f"-d {tag}")
        input("Press enter to continue after editing the changelog...")
        git.add_files([self.project.changelog_path])
        git.commit_files([self.project.changelog_path], "chore: update changelog")
        with git.capturing_output():
            git.tag(tag)
        # Sync with remote
        sync = f"origin {git.current_branch} --tags"
        git.pull(sync)
        git.push(sync)
        if args.publish:
            self.do_publish()
        if args.install:
            pip.main(["install", "."])


def main():
    """ """
    command = "" if len(sys.argv) < 2 else sys.argv[1]
    shell = HassleShell(command)
    if command == "help" and len(sys.argv) == 3:
        input_ = f"help {sys.argv[2]}"
    # Doing this so args that are multi-word strings don't get interpreted as separate args.
    elif command:
        input_ = f"{command} " + " ".join([f'"{arg}"' for arg in sys.argv[2:]])
    else:
        input_ = "help"
    shell.onecmd(input_)


if __name__ == "__main__":
    main()
